import textwrap

from arlunio.cli import Command, _cliCommand, _cliOption


class CommandSummaryOnly(Command):
    """This is an example command."""

    def run(self, reset: bool = False, verbose: bool = True, count: int = 1):
        pass


class CommandWithNoParams(Command):
    """This is an example command.

    Commands can contain an extended description that explains what the command
    does, how to invoke it as well as any gotchas or noteable command line
    arguments.
    """

    def run(self, reset: bool = False, verbose: bool = True, count: int = 1):
        pass


class CommandWithParams(Command):
    """This is an example command.

    Commands can contain an extended description that explains what the command
    does, how to invoke it as well as any gotchas or noteable command line
    arguments.

    :param rEset: Reset the thing back to its default state
    :param verbo3e: Use this if you want to gain further insight into what is
                    happening with the steps.
    :param num_tri8s:Repeat the command <N> times.

    """

    def run(self, rEset: bool = False, verbo3e: bool = True, num_tri8s: int = 1):
        pass


class TestCliCommand:
    """Tests for our docstring parsing function."""

    def test_construct_only_summary(self):
        """Ensure that we can parse the relevant information from a command that only
        provides a summary."""

        command = _cliCommand.from_command("example", CommandSummaryOnly)

        assert command.name == "example"
        assert command.summary == "This is an example command."
        assert command.description is None

        expected_options = {
            "--reset": _cliOption(
                name="--reset", argparse_args={"action": "store_true"}
            ),
            "--verbose": _cliOption(
                name="--verbose", argparse_args={"action": "store_false"}
            ),
            "--count": _cliOption(
                name="--count", argparse_args={"default": 1, "type": int}
            ),
        }

        actual_options = {opt.name: opt for opt in command.options}
        assert actual_options == expected_options

    def test_construct_with_description(self):
        """Ensure that we can parse the relevant information from a command that only
        gives a description."""

        command = _cliCommand.from_command("example", CommandWithNoParams)

        assert command.name == "example"
        assert command.summary == "This is an example command."

        expected_description = """\
        Commands can contain an extended description that explains what the command
        does, how to invoke it as well as any gotchas or noteable command line
        arguments."""

        assert command.description == textwrap.dedent(expected_description)

        expected_options = {
            "--reset": _cliOption(
                name="--reset", argparse_args={"action": "store_true"}
            ),
            "--verbose": _cliOption(
                name="--verbose", argparse_args={"action": "store_false"}
            ),
            "--count": _cliOption(
                name="--count", argparse_args={"default": 1, "type": int}
            ),
        }

        actual_options = {opt.name: opt for opt in command.options}
        assert actual_options == expected_options

    def test_contruct_full_command(self):
        """Ensure that we can parse the relevant information from a "full" command
        definition."""

        command = _cliCommand.from_command("example", CommandWithParams)

        assert command.name == "example"
        assert command.summary == "This is an example command."

        expected_description = """\
        Commands can contain an extended description that explains what the command
        does, how to invoke it as well as any gotchas or noteable command line
        arguments."""

        assert command.description == textwrap.dedent(expected_description)
        actual_options = {opt.name: opt for opt in command.options}

        expected_options = {
            "--rEset": _cliOption(
                name="--rEset",
                argparse_args={"action": "store_true"},
                description="Reset the thing back to its default state",
            ),
            "--verbo3e": _cliOption(
                name="--verbo3e",
                argparse_args={"action": "store_false"},
                description=(
                    "Use this if you want to gain further insight into what"
                    " is happening with the steps."
                ),
            ),
            "--num_tri8s": _cliOption(
                name="--num_tri8s",
                argparse_args={"default": 1, "type": int},
                description="Repeat the command <N> times.",
            ),
        }
        assert actual_options == expected_options



import arlunio as ar
import numpy as np


@ar.shape
def Circle(x, y, *, x0=0, y0=0, r=0.8, pt=None):
    """We define a circle using the following inequality.

    .. math::

       \\sqrt{(x - x_0)^2 + (y - y_0)^2} < r^2

    where:

    - :math:`(x_0, y_0)`: Defines the centre
    - :math:`r`: Controls the radius

    In python these variables can be set using the :code:`x0`, :code:`y0`
    and :code:`r` keyword arguments to the shape's constructor.

    By default when drawn the shape will draw a filled in circle, however the
    :code:`pt` keyword argument can be used to override this. If not :code:`None`
    the value of this argument will be used to control the thickness of the line
    used to draw the circle.
    """

    xc = x - x0
    yc = y - y0
    circle = np.sqrt(xc * xc + yc * yc)

    if pt is None:
        return circle < r * r

    r, R = (r - pt) ** 2, (r + pt) ** 2

    return np.logical_and(r < circle, circle < R)


@ar.shape
def Ellipse(x, y, *, x0=0, y0=0, a=2, b=1, r=0.8, pt=None):
    """An ellipse can be defined using the following inequality.

    .. math::

       \\frac{(x - x_0)^2}{a^2} + \\frac{(y - y_0)^2}{b^2} < r^2

    where:

    - :math:`(x_0, y_0)`: Defines the centre
    - :math:`r`: Controls the overall size
    - :math:`a`: Controls the width
    - :math:`b`: Controls the height

    In python these variables can be set using the :code:`x0`, :code:`y0`,
    :code:`r`, :code:`a` and :code:`b` keyword arguments to the shape's
    constructor.

    Increasing the value of :code:`a` will have the effect of stretching out the
    ellpise width-wise. Similarly increasing the value of :code:`b` will do the
    same but for the height. Whenever :code:`a == b` the result will be a circle.

    By default when drawn the shape will draw the ellipse filled in, this behavior
    can be changed using the :code:`pt` keyword argument. If not :code:`None` the
    value of this argument will be used as the thickness of the line used to draw
    the ellipse.
    """

    xc = (x - x0) ** 2
    yc = (y - y0) ** 2

    A = a ** 2
    B = b ** 2

    ellipse = np.sqrt(xc / A + yc / B)

    if pt is None:
        return ellipse < r * r

    r, R = (r - pt / 2) ** 2, (r + pt / 2) ** 2

    return np.logical_and(r < ellipse, ellipse < R)


@ar.shape
def Square(x, y, *, x0=0, y0=0, size=0.8):
    """A square."""

    xc = x - x0
    yc = y - y0

    size = size / 2

    return np.logical_and(np.abs(xc) < size, np.abs(yc) < size)



import argparse
import inspect
import logging
import re
import typing as t

import attr

logger = logging.getLogger(__name__)


class Command:
    """A command on the command line."""


@attr.s(auto_attribs=True)
class _cliOption:
    """Class that represents a cli option and relevant parameters."""

    name: str
    description: str = None
    argparse_args: t.Dict[str, t.Any] = attr.Factory(dict)

    @classmethod
    def from_parameter(cls, parameter: inspect.Parameter):
        params = {"name": "--" + parameter.name}
        argparse_args = {}

        if parameter.annotation == bool:
            if parameter.default:
                argparse_args["action"] = "store_false"
            else:
                argparse_args["action"] = "store_true"

        else:
            argparse_args["default"] = parameter.default
            argparse_args["type"] = parameter.annotation

        params["argparse_args"] = argparse_args
        return cls(**params)


def parse_options(method: callable) -> t.Dict[str, _cliOption]:
    """Given a method that implements a command, extract the information required to
    expose its arguments on the command line."""

    options = {}

    for param in inspect.signature(method).parameters.values():

        if param.name == "self":
            continue

        options[param.name] = _cliOption.from_parameter(param)

    return options


@attr.s(auto_attribs=True)
class _cliCommand:
    """Class that represents a command in a way that makes it easy for us to
    generate the command line interface from it."""

    name: str
    """The name of the command, as exposed on the command line."""

    summary: str = None
    """A short description of the command, as shown in the command list."""

    description: str = None
    """A more detailed description of the command, as shown in its dedicated
    help text"""

    options: t.List[_cliOption] = attr.Factory(list)
    """A list of command line options, with the information required to expose them
    via argparse."""

    @classmethod
    def from_command(cls, name: str, cmd: Command):
        # It 'works', it's not pretty, I don't like it
        params = {"name": name, "options": parse_options(getattr(cmd, "run"))}
        docstring = inspect.getdoc(cmd)

        param_tag = re.compile(":param ([a-zA-Z0-9_]+): ?([\\s\\w><.]+):?")

        try:

            summary_end = docstring.index("\n\n")
        except ValueError:
            params["summary"] = docstring
            params["options"] = list(params["options"].values())
            return cls(**params)

        params["summary"] = docstring[:summary_end]

        # Take into account the fact that index returns the index at the
        # start of the '\n\n' patternd
        description_start = summary_end
        next_parameter = param_tag.search(docstring[description_start:])

        if next_parameter is None:
            description = docstring[description_start:]
            params["description"] = description.lstrip().rstrip()
            params["options"] = list(params["options"].values())
            return cls(**params)

        description_end = description_start + next_parameter.span()[0]
        params["description"] = (
            docstring[description_start:description_end].rstrip().lstrip()
        )
        start = description_start

        while next_parameter is not None:
            name, helptext = next_parameter.groups()

            # Clean the text
            helptext = helptext.replace("\n", "")
            helptext = re.sub("[ ]+", " ", helptext)
            params["options"][name].description = helptext

            start += next_parameter.span()[1] - 1
            next_parameter = param_tag.search(docstring[start - 1 :])

        params["options"] = list(params["options"].values())
        return cls(**params)


def construct_parser(
    name: str, cmd: Command, subparser: argparse._SubParsersAction
) -> None:

    """Given a command instance and an argument parser, constuct the command line
    interface.

    :param name: The name of the command as it should be exposed on the command line.
    :param cmd: The class that represents the implementation of the command
    :param subparser: The subparsers object we can use to create our command with.
    """

    command = _cliCommand.from_command(name, cmd)
    parser = subparser.add_parser(
        command.name,
        help=command.summary,
        description=command.description,
        formatter_class=argparse.RawDescriptionHelpFormatter,
    )

    option_names = [opt.name[2:] for opt in command.options]

    for option in command.options:
        parser.add_argument(
            option.name, help=option.description, **option.argparse_args
        )

    def run(args: argparse.Namespace):
        params = {opt: getattr(args, opt) for opt in option_names}
        logger.debug(f"Parsed arguments: {params}")

        command = cmd()
        command.run(**params)

    parser.set_defaults(run=run)



import logging
import subprocess
import sys

import arlunio.cli as cli

logger = logging.getLogger(__name__)


class Repl(cli.Command):
    """Launch an interactive prompt."""

    def run(self):
        subprocess.run([sys.executable])



import argparse
import logging
import sys

import pkg_resources

import arlunio.cli

from ._version import __version__

logger = logging.getLogger(__name__)

LOG_LEVELS = [
    (logging.INFO, "%(message)s"),
    (logging.DEBUG, "[%(levelname)s][%(name)s]: %(message)s"),
]


def init_logging(verbose: int, quiet: bool) -> None:
    """Configure the logging system based on the given verbosity."""

    if quiet:
        return

    verbose = 0 if verbose < 0 else verbose

    try:
        level, fmt = LOG_LEVELS[verbose]
        others = False
    except IndexError:
        level, fmt = LOG_LEVELS[-1]
        others = True

    root = logging.getLogger()
    root.setLevel(level)

    console = logging.StreamHandler()
    console.setFormatter(logging.Formatter(fmt))

    if not others:
        console.addFilter(logging.Filter("arlunio"))

    root.addHandler(console)


def show_version():
    print(f"arlunio v{__version__}")


# Top level command line parser for arlunio
cli = argparse.ArgumentParser(description="Command line interface to arlunio")
commands = cli.add_subparsers(title="commands")

# "Global" arguments that can apply to any command.
cli.add_argument(
    "-q", "--quiet", help="disable all console output", action="store_true"
)
cli.add_argument(
    "-V", "--version", help="show version info and exit", action="store_true"
)
cli.add_argument(
    "-v",
    "--verbose",
    action="count",
    default=0,
    help="increase the verbosity of the output, can be repeated e.g. -v, -vv, ...",
)

# Register all the defined commands.
for cmd in pkg_resources.iter_entry_points("arlunio.cli.commands"):
    command = cmd.load()
    arlunio.cli.construct_parser(cmd.name, command, commands)


def main():
    args = cli.parse_args()

    if args.version:
        show_version()
        sys.exit(0)

    init_logging(args.verbose, args.quiet)

    if hasattr(args, "run"):
        args.run(args)
        return

    cli.print_help()


if __name__ == "__main__":
    main()
